#ifndef __TMPLNG_HPP__
#define __TMPLNG_HPP__

enum
{
	MOk,
	MCancel,
	MError,
	MWarning,

	MTempPanel,
	MTempPanelTitleNum,

	MDiskMenuString,

	MF7,
	MAltShiftF12,
	MAltShiftF2,
	MAltShiftF3,

	MTempUpdate,
	MTempSendFiles,

	MSwitchMenuTxt,
	MSwitchMenuTitle,

	MConfigTitle,
	MConfigAddToDisksMenu,
	MConfigDisksMenuDigit,
	MConfigAddToPluginsMenu,

	MConfigCommonPanel,
	MSafeModePanel,
	MAnyInPanel,

	MReplaceInFilelist,
	MMenuForFilelist,
	MCopyContents,

	MFullScreenPanel,

	MColumnTypes,
	MColumnWidths,
	MStatusColumnTypes,
	MStatusColumnWidths,

	MMask,
	MPrefix,

	MConfigNewOption,
	MNewPanelForSearchResults,
	MListFilePath,

	MCopyContentsMsg
};

#endif	/* __TMPLNG_HPP__ */
