enum
{
	MFileCase,

	MOk,
	MCancel,
	MReset,

	MName,
	MExtension,

	MLower,
	MUpper,
	MFirst,
	MTitle,
	MNone,
	MLowerExt,
	MUpperExt,
	MFirstExt,
	MTitleExt,
	MNoneExt,

	MSkipMixedCase,
	MProcessSubDir,
	MProcessDir,
	MWordDiv,

	MConverting,
	MCurRun,

};
