#ifndef GUIDS_HPP_BC5198A8_B680_42BC_8A7A_667ED84D79A4
#define GUIDS_HPP_BC5198A8_B680_42BC_8A7A_667ED84D79A4
#pragma once

extern INT_PTR g_plugin_id;

#ifndef Z7_DEFINE_GUID

// get rid of dancing with a tambourine (aka INITGUID)
#define Z7_DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8)                                      \
	constexpr GUID name = {                                                                                  \
			l, w1, w2, {b1, b2, b3, b4, b5, b6, b7, b8} \
	   }
#endif

Z7_DEFINE_GUID(c_plugin_guid, 0x65642111, 0xAA69, 0x4B84, 0xB4, 0xB8, 0x92, 0x49, 0x57, 0x9E, 0xC4, 0xFA);
Z7_DEFINE_GUID(c_plugin_menu_guid, 0x7BCFBA0E, 0x4EF5, 0x466D, 0xB5, 0xB9, 0x23, 0x52, 0x34, 0x59, 0xD6,
		0xAF);
Z7_DEFINE_GUID(c_plugin_config_guid, 0xEE3D5A10, 0x3D86, 0x4C8C, 0xB2, 0x6C, 0x88, 0x87, 0xD9, 0xE0, 0x7E,
		0xBE);
Z7_DEFINE_GUID(c_password_dialog_guid, 0x761F3B4C, 0xFC45, 0x4A9D, 0xA3, 0x83, 0x3F, 0x75, 0xD5, 0x05, 0xA4,
		0x3B);
Z7_DEFINE_GUID(c_overwrite_dialog_guid, 0x83B02899, 0x4590, 0x47F9, 0xB4, 0xC1, 0x6B, 0xC6, 0x6C, 0x6C, 0xA4,
		0xF9);
Z7_DEFINE_GUID(c_extract_dialog_guid, 0x97877FD0, 0x78E6, 0x4169, 0xB4, 0xFB, 0xD7, 0x67, 0x46, 0x24, 0x9F,
		0x4D);
Z7_DEFINE_GUID(c_update_dialog_guid, 0xCD57D7FA, 0x552C, 0x4E31, 0x8F, 0xA8, 0x73, 0xD9, 0x70, 0x4F, 0x06,
		0x66);
Z7_DEFINE_GUID(c_settings_dialog_guid, 0x08A1229B, 0xAD54, 0x451B, 0x8B, 0x53, 0x6D, 0x5F, 0xD3, 0x5B, 0xCF,
		0xAA);
Z7_DEFINE_GUID(c_attr_dialog_guid, 0x17D10388, 0x46C7, 0x4BBD, 0xA3, 0xCA, 0x9A, 0x97, 0x77, 0x11, 0xF4,
		0xE5);
Z7_DEFINE_GUID(c_sfx_options_dialog_guid, 0x0DCE48E5, 0xB205, 0x44A0, 0xB8, 0xBF, 0x96, 0xB2, 0x8E, 0x2F,
		0xD3, 0xB3);
Z7_DEFINE_GUID(c_export_options_dialog_guid, 0x0DCE4899, 0xB205, 0x44A0, 0xB8, 0xBF, 0x96, 0xB2, 0x8E, 0x2F,
		0x99, 0xB3);
Z7_DEFINE_GUID(c_delete_files_dialog_guid, 0x4FCA2D71, 0x682C, 0x442F, 0x86, 0x35, 0x12, 0x11, 0xBD, 0x44,
		0x95, 0xFB);
Z7_DEFINE_GUID(c_overwrite_archive_dialog_guid, 0xE17D3FE5, 0xFA3D, 0x4A89, 0xBD, 0x14, 0xDE, 0x8B, 0x50,
		0x12, 0x2D, 0xD8);
Z7_DEFINE_GUID(c_progress_dialog_guid, 0x1C5D302E, 0x4AB7, 0x4402, 0x94, 0x05, 0xCB, 0xCE, 0xC9, 0x87, 0xA2,
		0x85);
Z7_DEFINE_GUID(c_interrupt_dialog_guid, 0x1132D7C0, 0x8137, 0x4503, 0x90, 0xE4, 0x04, 0xCC, 0x59, 0xE2, 0x9C,
		0x1C);
Z7_DEFINE_GUID(c_retry_ignore_dialog_guid, 0x576BC2CC, 0x5547, 0x421C, 0xAB, 0xF8, 0xB6, 0xDF, 0x74, 0x7E,
		0x71, 0x1C);
Z7_DEFINE_GUID(c_error_log_dialog_guid, 0x53FB8447, 0x1C00, 0x42CB, 0x93, 0x4C, 0x9A, 0xB2, 0xB5, 0x0A, 0x29,
		0xA9);
Z7_DEFINE_GUID(c_delete_profile_dialog_guid, 0x0A5411E3, 0x40F3, 0x415F, 0x81, 0x00, 0xAD, 0x45, 0xE6, 0x35,
		0xCB, 0xD1);
Z7_DEFINE_GUID(c_test_ok_dialog_guid, 0x2AE9F593, 0x46B0, 0x41AE, 0xAF, 0xF2, 0x29, 0xBA, 0x0C, 0x45, 0x88,
		0x32);
Z7_DEFINE_GUID(c_extract_params_saved_dialog_guid, 0x2B27B493, 0xE3D0, 0x445F, 0xAA, 0x0B, 0xFD, 0xF5, 0x1D,
		0x1D, 0xDC, 0x23);
Z7_DEFINE_GUID(c_arc_path_eval_dialog_guid, 0x6DA788EA, 0x813C, 0x46A2, 0x9E, 0x5A, 0xE1, 0x04, 0x73, 0x77,
		0x09, 0x48);
Z7_DEFINE_GUID(c_error_dialog_guid, 0x7C5D2FE2, 0x14B8, 0x47EF, 0xBF, 0x35, 0xD3, 0x14, 0xE1, 0xE2, 0xFA,
		0xB4);
Z7_DEFINE_GUID(c_format_menu_guid, 0x42C8F05D, 0x4E95, 0x4069, 0xA9, 0x84, 0x0C, 0xFD, 0x59, 0xE1, 0xB4,
		0x42);
Z7_DEFINE_GUID(c_main_menu_guid, 0xEAABFBA1, 0xD321, 0x4D8C, 0x95, 0x73, 0xA2, 0xA0, 0x6E, 0xC2, 0x26, 0x84);
Z7_DEFINE_GUID(c_arccmd_menu_guid, 0x83AD575C, 0xBBE4, 0x45A1, 0xA0, 0x33, 0x34, 0x5A, 0x4D, 0xFB, 0x6B,
		0x74);
Z7_DEFINE_GUID(c_create_dir_dialog_guid, 0x65593288, 0x4F2A, 0x4410, 0x9E, 0x3A, 0x7C, 0x43, 0x39, 0x81, 0x69,
		0x92);
Z7_DEFINE_GUID(c_save_profile_dialog_guid, 0x9B55A916, 0x1599, 0x4181, 0x90, 0xA5, 0xCA, 0x37, 0xDE, 0x34,
		0x57, 0x16);
Z7_DEFINE_GUID(c_generic_guid, 0x4756DC8E, 0x15E2, 0x41FA, 0xAE, 0x6D, 0xE9, 0x72, 0xB5, 0xFB, 0x43, 0xD8);
Z7_DEFINE_GUID(c_update_params_saved_dialog_guid, 0x1C453287, 0x733E, 0x4B3E, 0xB0, 0xE4, 0xCE, 0x0D, 0x8F,
		0xA9, 0x7D, 0xB6);
Z7_DEFINE_GUID(c_multi_select_dialog_guid, 0xBCBC467B, 0x4F16, 0x4CF1, 0x98, 0x95, 0x92, 0xA6, 0x16, 0xA9,
		0xBA, 0xBD);
Z7_DEFINE_GUID(c_format_library_info_dialog_guid, 0x223C2003, 0xA7FF, 0x4907, 0xA4, 0xA3, 0x6B, 0xF1, 0x0D,
		0xEA, 0x84, 0x32);
Z7_DEFINE_GUID(c_far_guid, 0x00000000, 0x0000, 0x0000, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00);


#endif	  // GUIDS_HPP_BC5198A8_B680_42BC_8A7A_667ED84D79A4
