ln -s ../far2sdk/farcolor.h farcolor.hpp
ln -s ../far2sdk/farkeys.h farkeys.hpp
ln -s ../far2sdk/farplug-mb.h pluginold.hpp
ln -s ../far2sdk/farplug-wide.h plugin.hpp
ln -s ../far2sdk/farcommon.h farcommon.hpp
