#ifndef COLORER_STRINGS_H
#define COLORER_STRINGS_H

#include "colorer/strings/legacy/common_legacy.h"
#include "colorer/strings/legacy/Character.h"
#include "colorer/strings/legacy/Encodings.h"
#include "colorer/strings/legacy/UStr.h"
#include "colorer/strings/legacy/UnicodeStringContainer.h"
#include "colorer/strings/legacy/UnicodeTools.h"

#endif  // COLORER_STRINGS_H
