/*
    This file was automatically generated from
    UNICODE 3.2.0 UnicodeData.txt base.
    Do not edit it with hands.
*/

// ansi class names
static char char_category_names[][3] = { "Cn", "Cc", "Cf", "Cs", "Ll", "Lm", "Lo", "Lt", "Lu", "Mc", "Me", "Mn", "Nd", "Nl", "No", "Pc", "Pd", "Pe", "Pf", "Pi", "Po", "Ps", "Sc", "Sk", "Sm", "So", "Zl", "Zp", "Zs",  };
/*
// ansi class names
static char char_bidi_names[][4] = { "AL", "AN", "B", "BN", "CS", "EN", "ES", "ET", "L", "LRE", "LRO", "NSM", "ON", "PDF", "R", "RLE", "RLO", "S", "WS",  };


// unc class names
static wchar wchar_category_names[][3] = { L"Cn", L"Cc", L"Cf", L"Cs", L"Ll", L"Lm", L"Lo", L"Lt", L"Lu", L"Mc", L"Me", L"Mn", L"Nd", L"Nl", L"No", L"Pc", L"Pd", L"Pe", L"Pf", L"Pi", L"Po", L"Ps", L"Sc", L"Sk", L"Sm", L"So", L"Zl", L"Zp", L"Zs",  };

// unc class names
static wchar wchar_bidi_names[][4] = { L"AL", L"AN", L"B", L"BN", L"CS", L"EN", L"ES", L"ET", L"L", L"LRE", L"LRO", L"NSM", L"ON", L"PDF", L"R", L"RLE", L"RLO", L"S", L"WS",  };
*/



