#include "version.hpp"

#define PLUGIN_MAJOR 1
#define PLUGIN_MINOR 9
#define PLUGIN_DESC "Directory listing parcer FTP plugin"
#define PLUGIN_NAME "ftpDirList"
#define PLUGIN_FILENAME "ftpDirList.fll"

ftpgenericpluginrc(PLUGIN_MAJOR, PLUGIN_MINOR, PLUGIN_DESC, PLUGIN_NAME, PLUGIN_FILENAME)
