#ifndef _CALC_MESSAGES_H_
#define _CALC_MESSAGES_H_

enum
{
	mName, mMenuName, mConfigName, mLang, mBottomHint,
	
	mDialogs, 
	mOk, mSave, mCancel, mAdvanced,

	mConfigShowResults,
	mConfigCaseSensitive,
	mConfigShowPeriodic,
	mConfigMaxPeriod,
	mConfigPadZeroes,
	mConfigRightAligned,
	mConfigHistory,
	mConfigHide,
	mConfigShowAbove,
	mConfigShowBelow,
	mConfigLines,
	mConfigAutocomplete,
	mConfigDelimiters,
	mConfigDecimalSymbol,
	mConfigUseRegional,
	mConfigArguments,
	mConfigDigitDelimiter,
	
	mNoError,
	mError1, mError2, mError3, mError4, mError5, mError6, mError7, mError8,
};


#endif
