#include <all_far.h>

#include "Int.h"

#define LLog(v)                                                                                                \
	if (FP_Info)                                                                                               \
	Log(v)

#ifdef _DEBUG
void ShowMemInfo(void) {}
#endif	// BC
