/*
FAR manager incremental search plugin, search as you type in editor.
Copyright (C) 1999-2019, Stanislav V. Mekhanoshin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#pragma once

enum
{
	MIncSearch,

	MOk,
	MCancel,

	MSearchForward,
	MSearchBackward,
	MFindNext,
	MFindPrevious,
	MSetup,

	MCaseSensitive,
	MKeepSelection,
	MBeepOnMismatch,
	MRestartEOF,
	MUseSelection,
	MAutoNext,
	MBSunroll,

	MEOF,
	MBOF,
	MNotFound
};
