// #include <farversion.hpp>

#define PLUGIN_BUILD 348
#define PLUGIN_DESC		L"Archive support for far2l fork of Far Manager (based on 7-Zip project)"
#define PLUGIN_NAME		L"ArcLite"
#define PLUGIN_FILENAME L"arclite.far-plug-wide"
#define PLUGIN_AUTHOR	FARCOMPANYNAME
// #define PLUGIN_VERSION
// MAKEFARVERSION(FARMANAGERVERSION_MAJOR,FARMANAGERVERSION_MINOR,FARMANAGERVERSION_REVISION,PLUGIN_BUILD,VS_RELEASE)
#define PLUGIN_VERSION	MAKEFARVERSION(FARMANAGERVERSION_MAJOR, FARMANAGERVERSION_MINOR)

// #define FARMANAGERVERSION_MAJOR 2
// #define FARMANAGERVERSION_MINOR 6
