#pragma once

enum LanguageID
{
	MPluginTitle,
	MChooseApplication,

	MOk,
	MCancel,
	MError,
	MNotRealNames,
	MSaveConfigError,
	MNoAppsFound,
	MAndMore,
	MCannotExecute,

	MConfigTitle,

	MUseExternalTerminal,
	MNoWaitForCommandCompletion,
	MClearSelection,
	MConfirmLaunchOption,

	MUseXdgMimeTool,
	MUseFileTool,
	MUseExtensionBasedFallback,
	MLoadMimeTypeAliases,
	MLoadMimeTypeSubclasses,
	MResolveStructuredSuffixes,
	MUseGenericMimeFallbacks,
	MShowUniversalHandlers,
	MUseMimeinfoCache,
	MFilterByShowIn,
	MValidateTryExec,
	MSortAlphabetically,
	MTreatUrlsAsPaths,

	MDetails,

	MPathname,
	MFilesSelected,
	MMimeType,
	MLaunchCommand,
	MClose,
	MLaunch,

	MDesktopFile,
	MSource,
	MFullScanFor,
	MFor,
	MIn,

	MAppName,
	MFullPath,
	MExecutableFile,
	MVersion,
	MBundleVersion,

	MConfirmLaunchTitle,
	MConfirmLaunchMessage
};
