#pragma once

void create_sfx_module(const std::wstring &file_path, const struct SfxOptions &sfx_options);
void attach_sfx_module(const std::wstring &file_path, const struct SfxOptions &sfx_options);
