#ifndef COLORER_COMMONSTRING_H
#define COLORER_COMMONSTRING_H

/// default unicode char definition

using UChar = char16_t;
using wchar = wchar_t;
using w2char = char16_t;
using w4char = char32_t;
using byte = unsigned char;

constexpr UChar BAD_WCHAR = 0xFFFF;

#endif // COLORER_COMMONSTRING_H


