#pragma once
enum {
M_TITLE,

M_OK,
M_CANCEL,

M_HINT_SHIFT_TAB,
M_HINT_TAB,
M_TEXT_AUTOENABLE,
M_TEXT_ENABLED,

M_TEXT_MAX_LINE_DELTA,
M_TEXT_MAX_LINE_LENGTH,
M_TEXT_MAX_WORD_LENGTH,
M_TEXT_MIN_PREFIX_LENGTH,

};
