#pragma once
BOOL WINAPI _export SEVENZ_IsArchive(const char *Name,const unsigned char *Data,int DataSize);
BOOL WINAPI _export SEVENZ_OpenArchive(const char *Name,int *Type,bool Silent);
int WINAPI _export SEVENZ_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export SEVENZ_CloseArchive(struct ArcInfo *Info);
BOOL WINAPI _export SEVENZ_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export SEVENZ_GetDefaultCommands(int Type,int Command,std::string &Dest);
BOOL WINAPI _export CAB_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export CAB_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export CAB_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export CAB_CloseArchive(struct ArcInfo *Info);
DWORD WINAPI _export CAB_GetSFXPos(void);
BOOL WINAPI _export CAB_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export CAB_GetDefaultCommands(int Type, int Command, std::string &Dest);
void WINAPI _export RAR_SetFarInfo(const struct PluginStartupInfo *Info);
BOOL WINAPI _export RAR_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export RAR_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export RAR_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export RAR_CloseArchive(struct ArcInfo *Info);
DWORD WINAPI _export RAR_GetSFXPos(void);
BOOL WINAPI _export RAR_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export RAR_GetDefaultCommands(int Type, int Command, std::string &Dest);
void WINAPI _export CUSTOM_SetFarInfo(const struct PluginStartupInfo *Info);
DWORD WINAPI _export CUSTOM_LoadFormatModule(const char *ModuleName);
BOOL WINAPI _export CUSTOM_IsArchive(const char *FName, const unsigned char *Data, int DataSize);
DWORD WINAPI _export CUSTOM_GetSFXPos(void);
BOOL WINAPI _export CUSTOM_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export CUSTOM_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export CUSTOM_CloseArchive(struct ArcInfo *Info);
BOOL WINAPI _export CUSTOM_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export CUSTOM_GetDefaultCommands(int Type, int Command, std::string &Dest);
void WINAPI _export TARGZ_SetFarInfo(const struct PluginStartupInfo *Info);
BOOL WINAPI _export TARGZ_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export TARGZ_OpenArchive(const char *Name, int *Type, bool Silent);
DWORD WINAPI _export TARGZ_GetSFXPos(void);
int WINAPI _export TARGZ_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export TARGZ_CloseArchive(struct ArcInfo *Info);
BOOL WINAPI _export TARGZ_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export TARGZ_GetDefaultCommands(int Type, int Command, std::string &Dest);
BOOL WINAPI _export ARC_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
DWORD WINAPI _export ARC_GetSFXPos(void);
BOOL WINAPI _export ARC_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export ARC_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export ARC_CloseArchive(struct ArcInfo *Info);
BOOL WINAPI _export ARC_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export ARC_GetDefaultCommands(int Type, int Command, std::string &Dest);
BOOL WINAPI _export LIBARCH_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export LIBARCH_OpenArchive(const char *Name, int *Type, bool Silent);
DWORD WINAPI _export LIBARCH_GetSFXPos(void);
int WINAPI _export LIBARCH_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export LIBARCH_CloseArchive(struct ArcInfo *Info);
BOOL WINAPI _export LIBARCH_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export LIBARCH_GetDefaultCommands(int Type, int Command, std::string &Dest);
BOOL WINAPI _export ACE_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export ACE_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export ACE_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export ACE_CloseArchive(struct ArcInfo *Info);
DWORD WINAPI _export ACE_GetSFXPos(void);
BOOL WINAPI _export ACE_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export ACE_GetDefaultCommands(int Type, int Command, std::string &Dest);
void  WINAPI _export HA_SetFarInfo(const struct PluginStartupInfo *Info);
BOOL WINAPI _export HA_IsArchive(const char *Name,const unsigned char *Data,int DataSize);
BOOL WINAPI _export HA_OpenArchive(const char *Name,int *Type,bool Silent);
int WINAPI _export HA_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export HA_CloseArchive(struct ArcInfo *Info);
BOOL WINAPI _export HA_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export HA_GetDefaultCommands(int Type,int Command,std::string &Dest);
BOOL WINAPI _export ZIP_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export ZIP_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export ZIP_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export ZIP_CloseArchive(struct ArcInfo *Info);
DWORD WINAPI _export ZIP_GetSFXPos(void);
BOOL WINAPI _export ZIP_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export ZIP_GetDefaultCommands(int Type, int Command, std::string &Dest);
BOOL WINAPI _export ARJ_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export ARJ_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export ARJ_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export ARJ_CloseArchive(struct ArcInfo *Info);
DWORD WINAPI _export ARJ_GetSFXPos(void);
BOOL WINAPI _export ARJ_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export ARJ_GetDefaultCommands(int Type, int Command, std::string &Dest);
void WINAPI _export LZH_SetFarInfo(const struct PluginStartupInfo *Info);
BOOL WINAPI _export LZH_IsArchive(const char *Name, const unsigned char *Data, int DataSize);
BOOL WINAPI _export LZH_OpenArchive(const char *Name, int *Type, bool Silent);
int WINAPI _export LZH_GetArcItem(struct ArcItemInfo *Info);
BOOL WINAPI _export LZH_CloseArchive(struct ArcInfo *Info);
DWORD WINAPI _export LZH_GetSFXPos(void);
BOOL WINAPI _export LZH_GetFormatName(int Type, std::string &FormatName, std::string &DefaultExt);
BOOL WINAPI _export LZH_GetDefaultCommands(int Type, int Command, std::string &Dest);
