/*
FAR manager incremental search plugin, search as you type in editor.
Copyright (C) 1999-2019, Stanislav V. Mekhanoshin

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#pragma once

#define NOANIMATE
#define NOAVICAP
#define NOAVIFILE
#define NOAVIFMT
#define NOBITMAP
// #define NOCLIPBOARD
#define NOCOLOR
#define NOCOMPMAN
#define NOCRYPT
#define NOCTLMGR
#define NODATETIMEPICK
#define NODDEMLSPY
#define NODEFERWINDOWPOS
#define NODESKTOP
#define NODRAGLIST
#define NODRAWDIB
#define NODRAWTEXT
#define NOEXCHEXTGUIDS
#define NOEXCHFORMGUIDS
#define NOEXTAPI
#define NOFLATSBAPIS
#define NOFONTSIG
#define NOGDI
#define NOGDICAPMASKS
#define NOHEADER
#define NOHELP
#define NOHOTKEY
#define NOICONS
#define NOIDLEENGINE
#define NOIMAGEAPIS
#define NOIME
#define NOIPADDRESS
#define NOJPEGDIB
#define NOKEYSTATES
#define NOLISTVIEW
#define NOMCIWND
#define NOMCX
#define NOMDI
#define NOMENUHELP
#define NOMENUS
#define NOMETAFILE
#define NOMINMAX
#define NOMMIDS
#define NOMMREG
#define NOMONTHCAL
#define NOMSACM
#define NOMSG
#define NONATIVEFONTCTL
#define NONCMESSAGES
#define NONEWIC
#define NONEWRIFF
#define NONEWWAVE
// #define NONLS
#define NOPAGESCROLLER
#define NOPENALC
#define NOPENAPIFUN
#define NOPENAPPS
#define NOPENBEDIT
#define NOPENBMP
#define NOPENCTL
#define NOPENCURS
#define NOPENDATA
#define NOPENDICT
#define NOPENDRIVER
#define NOPENHEDIT
#define NOPENHRC
#define NOPENIEDIT
#define NOPENINKPUT
#define NOPENMISC
#define NOPENMSGS
#define NOPENRC1
#define NOPENTARGET
#define NOPENVIRTEVENT
#define NOPROGRESS
#define NORASTEROPS
#define NOREBAR
#define NORESOURCE
#define NOSCROLL
#define NOSECURITY
#define NOSERVICE
#define NOSHLWAPI
#define NOSHOWWINDOW
#define NOSTATUSBAR
#define NOSYSCOMMANDS
#define NOSYSMETRICS
#define NOSYSPARAMSINFO
#define NOTABCONTROL
#define NOTEXTMETRIC
#define NOTOOLBAR
#define NOTOOLTIPS
#define NOTRACKBAR
#define NOTRACKMOUSEEVENT
#define NOTREEVIEW
#define NOUPDOWN
#define NOUSEREXCONTROLS
#define NOWH
#define NOWINABLE
#define NOWINDOWSTATION
#define NOWINMESSAGES
#define NOWINOFFSETS
#define NOWINSTYLES
#define NO_BASEINTERFACE_FUNCS
#define NO_INTSHCUT_GUIDS
#define NO_SHDOCVW_GUIDS
#define NO_SHLWAPI_GDI
#define NO_SHLWAPI_PATH
#define NO_SHLWAPI_REG
#define NO_SHLWAPI_STREAM
#define NO_SHLWAPI_STRFCNS
#include <time.h>
#include <windows.h>
