/*
    This file was automatically generated from
    UNICODE 3.2.0 UnicodeData.txt base.
    Do not edit it with hands.
*/

// all associations except CHAR_CATEGORY_Cn = 0
// could be changed with changes in UNICODE database
enum ECharCategory{
CHAR_CATEGORY_Cn = 0, // unassigned
CHAR_CATEGORY_Cc = 1,
CHAR_CATEGORY_Cf = 2,
CHAR_CATEGORY_Cs = 3,
CHAR_CATEGORY_Ll = 4,
CHAR_CATEGORY_Lm = 5,
CHAR_CATEGORY_Lo = 6,
CHAR_CATEGORY_Lt = 7,
CHAR_CATEGORY_Lu = 8,
CHAR_CATEGORY_Mc = 9,
CHAR_CATEGORY_Me = 10,
CHAR_CATEGORY_Mn = 11,
CHAR_CATEGORY_Nd = 12,
CHAR_CATEGORY_Nl = 13,
CHAR_CATEGORY_No = 14,
CHAR_CATEGORY_Pc = 15,
CHAR_CATEGORY_Pd = 16,
CHAR_CATEGORY_Pe = 17,
CHAR_CATEGORY_Pf = 18,
CHAR_CATEGORY_Pi = 19,
CHAR_CATEGORY_Po = 20,
CHAR_CATEGORY_Ps = 21,
CHAR_CATEGORY_Sc = 22,
CHAR_CATEGORY_Sk = 23,
CHAR_CATEGORY_Sm = 24,
CHAR_CATEGORY_So = 25,
CHAR_CATEGORY_Zl = 26,
CHAR_CATEGORY_Zp = 27,
CHAR_CATEGORY_Zs = 28,

CHAR_CATEGORY_LAST
};

enum ECharBidi{
BIDI_CLASS_AL = 0,
BIDI_CLASS_AN = 1,
BIDI_CLASS_B = 2,
BIDI_CLASS_BN = 3,
BIDI_CLASS_CS = 4,
BIDI_CLASS_EN = 5,
BIDI_CLASS_ES = 6,
BIDI_CLASS_ET = 7,
BIDI_CLASS_L = 8,
BIDI_CLASS_LRE = 9,
BIDI_CLASS_LRO = 10,
BIDI_CLASS_NSM = 11,
BIDI_CLASS_ON = 12,
BIDI_CLASS_PDF = 13,
BIDI_CLASS_R = 14,
BIDI_CLASS_RLE = 15,
BIDI_CLASS_RLO = 16,
BIDI_CLASS_S = 17,
BIDI_CLASS_WS = 18,

BIDI_CLASS_LAST
};





